#include "StdAfx.h"
#include "DesktopItem.h"
#include "DesktopTexture.h"
#include "Font.h"
#include "ContextMenu.h"

#define TIMER_MOUSETRACK 1

extern DesktopItemList desktopitemList;

extern Font *defaultfont;
extern DesktopTexture *defaultbackground;

DesktopItem::DesktopItem(const string &name):actualheight(0),actualwidth(0),actualx(0),actualy(0)
{
	this->name = name;

	hWnd = 0;
	hInstance = 0;

	TooltipHints = 0;

	backgroundDC = 0;
	backgroundBitmap = 0;
	
	mousePressed = false;
	mouseInside = false;

	iconOverlay = false;

	DIsnapto = false;

	visible = false;
}

DesktopItem::~DesktopItem()
{
	if (hWnd && TooltipHints)
	{
		removeHint(hWnd);
		DestroyWindow(TooltipHints);
		TooltipHints = 0;
	}

	if(IsWindow(hWnd))
	{
		hide();

		KillTimer(hWnd, TIMER_MOUSETRACK);

		DestroyWindow(hWnd);
	}

	if(backgroundBitmap)
	{
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
		DeleteObject(backgroundBitmap);
	}
	if (backgroundDC)
		DeleteDC(backgroundDC);
}

void DesktopItem::desktopload(HINSTANCE hInstance, string configLine)
{
	this->hInstance = hInstance;
	
	if (IsWindow(hWnd))
		return;

	hWnd = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xDesktopItem",
		name.c_str(),
		WS_POPUP,
		actualx, actualy,			
		actualwidth, actualheight,
		GetLitestepWnd(),  
		0,
		hInstance,
		this);

	if (hWnd)
	{
		SetWindowLong(hWnd, GWL_USERDATA, magicDWord);
		
		//ToolTips for Icons
		TooltipHints = CreateWindow(TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP,
									  CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
									  NULL, NULL, hInstance, NULL);
		if (TooltipHints)
		{
			SetWindowPos(TooltipHints, HWND_TOPMOST, 0, 0, 0, 0,
						 SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

			SendMessage(TooltipHints, TTM_SETMAXTIPWIDTH, 0, 600);
		}

		desktopreconfigure(configLine);
	}
	else
		delete this;
}

void DesktopItem::desktopreconfigure(string configLine)
{
		NameValuePair iconTextValues[] = {
			{ "top", 1 },
			{ "bottom", 2 },
			{ "left", 3 },
			{ "right", 4 },
			{ 0, 0 }
		};

		//Painting Background

		//Prepare Settings
		//----------------------------------------------------------------
		char xname[32], ix[16], iy[16], path[MAX_PATH_LENGTH];
		char *buffers[] = {xname, ix, iy, path};

		LCTokenize(configLine.c_str(), buffers, 4, NULL);
		
		actualx = atoi( ix );
		actualy = atoi( iy );
		actualwidth = GetxRCInt("xDesktop", "ItemWidth", 100);
		actualheight = GetxRCInt("xDesktop", "ItemHeight", 75);
		
		//----------------------------------------------------------------

		//SnapTo Settings
		char diX[64], diY[64], diWidth[64], diHeight[64], diDistance[64];
		char *DIbuffers[] = {diX, diY, diWidth, diHeight, diDistance};

		if ( LCTokenize( (GetxRCLine("xdesktop", "SnapTo", "")).c_str(), DIbuffers, 5, NULL) == 5 )
		{
			DIsnapto = true;
			DIx = atoi(diX);
			DIy = atoi(diY);
			DIwidth = atoi(diWidth);
			DIheight = atoi(diHeight);
			DIsnapdistance = atoi(diDistance);
		}

		iconCommand = path;		

		//Prepare Special DesktopSettings
		//----------------------------------------------------------------
		int iconSize = GetxRCInt("xdesktop", "IconSize", 32);
		int iconTextAlign = GetxRCNamedValue("xdesktop", "TextPosition", iconTextValues, 2);
		int iconTextSpacing = GetxRCInt("xdesktop" , "TextSpacing", 5);
		//----------------------------------------------------------------

		//Painting Text
		if (iconTextAlign == 1 || iconTextAlign == 2)
		{
			leftBorder = 0;
			rightBorder = 0;
			align = DT_CENTER;
		}
		else if (iconTextAlign == 3)
		{
			leftBorder = 0;
			rightBorder = iconSize+iconTextSpacing+1;
			align = DT_RIGHT;
		}
		else if (iconTextAlign == 4)
		{
			leftBorder = iconSize+iconTextSpacing+1;
			rightBorder = 0;
			align = DT_LEFT;
		}

		if (iconTextAlign == 1)
		{
			topBorder = 0;
			bottomBorder = iconSize+iconTextSpacing+1;
			vertAlign = ALIGN_BOTTOM;
		}
		else if (iconTextAlign == 2)
		{
			topBorder = iconSize+iconTextSpacing+1;
			bottomBorder = 0;
			vertAlign = ALIGN_TOP;
		}
		else if (iconTextAlign == 3 || iconTextAlign == 4)
		{
			topBorder = 0;
			bottomBorder = 0;
			vertAlign = ALIGN_CENTER;
		}

		bool itemIsFolder = false;
		
		HRESULT hres;
		IShellLink* psl;
		SHFILEINFO shFileInfo;

		char szGetPath[MAX_PATH_LENGTH];
		if (strstr(iconCommand.c_str(), "::{") != NULL)
		{
			itemIsFolder = true;
			SHGetFileInfo(iconCommand.c_str(), NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_DISPLAYNAME | SHGFI_TYPENAME);
			text = shFileInfo.szDisplayName;
			strcpy(tooltip , "Type:    ");
			strcat(tooltip , shFileInfo.szTypeName);
			strcat(tooltip, "\nTarget: ");
			strcat(tooltip , shFileInfo.szDisplayName);
		}
		else
		{
			SHGetFileInfo(iconCommand.c_str(), NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_DISPLAYNAME | SHGFI_TYPENAME | SHGFI_ATTRIBUTES);
			text = shFileInfo.szDisplayName;
			strcpy(tooltip , "Type:    ");
			strcat(tooltip , shFileInfo.szTypeName);
		
			if (shFileInfo.dwAttributes & SFGAO_FOLDER)
				itemIsFolder = true;
		}

		//Create the ShellLink object
		hres = CoCreateInstance(CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID*) &psl);

		if (SUCCEEDED(hres))
		{
			IPersistFile* ppf;
			//Bind the ShellLink object to the Persistent File
			hres = psl->QueryInterface( IID_IPersistFile, (LPVOID *) &ppf);

			if (SUCCEEDED(hres))
			{
				WORD wsz[MAX_PATH_LENGTH];
				//Get a UNICODE wide string wsz from the Link path
				MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, path, -1, wsz, MAX_PATH_LENGTH);
					
				//Read the link into the persistent file
				hres = ppf->Load(wsz, 0);
					
				if (SUCCEEDED(hres))
				{
					//Read the target information from the link object
					//UNC paths are supported (SLGP_UNCPRIORITY)
					psl->GetPath(szGetPath, MAX_PATH_LENGTH, NULL, SLGP_UNCPRIORITY);
						
					strcat(tooltip, "\nTarget: ");
					strcat(tooltip, szGetPath);

					psl->GetDescription(szGetPath, MAX_PATH_LENGTH);

					if (stricmp(szGetPath, "") != NULL)
					{
						strcat(tooltip, "\n\nDescription:\n");
						strcat(tooltip, szGetPath);

						SendMessage(TooltipHints, TTM_SETMAXTIPWIDTH, 0, 300);
					}
				}
				else if (strstr(path, "::{") == NULL)
				{
					strcat(tooltip, "\nName:   ");
					strcat(tooltip, iconCommand.c_str());
				}
			}
		}
		psl->Release();

		addHint(hWnd, tooltip);

		//Events/Actions
		easyDesktopFocus = GetxRCBool("xdesktop", "EasyFocus");
		
		leftClickCommand = GetxRCString("xdesktop", "OnLeftClick", ".move");
		leftDoubleClickCommand  = GetxRCString("xdesktop", "OnLeftDoubleClick", ".open");
		middleClickCommand = GetxRCString("xdesktop", "OnMiddleClick", ".properties");
		middleDoubleClickCommand = GetxRCString("xdesktop", "OnMiddleDoubleClick", ".none");
		rightClickCommand =  GetxRCString("xdesktop", "OnRightClick", ".context");
		rightDoubleClickCommand =  GetxRCString("xdesktop", "OnRightDoubleClick", ".none");

		if (itemIsFolder)
		{
			if (strstr(iconCommand.c_str(), "::{20D04FE0-3AEA-1069-A2D8-08002B30309D}") != NULL)
			{
				dropCommand = "";
				itemIsFolder = false;
			}
			else if (strstr(iconCommand.c_str(), "::{450D8FBA-AD25-11D0-98A8-0800361B1103}") != NULL)
			{
				SHGetSpecialFolderPathA(NULL, path, CSIDL_PERSONAL, false);
				dropCommand = "copy \"";
				dropCommand.append( path );
				dropCommand.append( "\"" );
			}
			else if (strstr(iconCommand.c_str(), "::{208D2C60-3AEA-1069-A2D7-08002B30309D}") != NULL)
			{
				dropCommand = "";
				itemIsFolder = false;
			}
			else if (strstr(iconCommand.c_str(), "::{645FF040-5081-101B-9F08-00AA002F954E}") != NULL)
			{
				dropCommand = "delete files";
			}
			else
				dropCommand = ("copy \"" + iconCommand + "\"");
		}
		else
		{
			dropCommand = "";
		}

		//Moving
		moveable = GetxRCBool("xdesktop", "Moveable", true);
		//Apply Selection of MoveButton
		if (stricmp(middleClickCommand.c_str(), ".move") == NULL)
			moveButton = 3;

		if (stricmp(rightClickCommand.c_str(), ".move") == NULL)
			moveButton = 2;

		if (stricmp(leftClickCommand.c_str(), ".move") == NULL)
			moveButton = 1;

		if (moveButton == 0)
			moveable = false;

		//Apply Selection of clicks
		if (stricmp(leftClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			leftClickCommand = ("!execute [\"" + iconCommand + "\"]");
		else if (stricmp(leftClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			leftClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else if (stricmp(leftClickCommand.c_str(), ".context") != NULL && stricmp(leftClickCommand.c_str(), ".properties") != NULL)
			leftClickCommand = "";

		if (stricmp(leftDoubleClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			leftDoubleClickCommand = ("!execute [\"" + iconCommand + "\"]");
		else if (stricmp(leftDoubleClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			leftDoubleClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else
			leftDoubleClickCommand = "";

		if (stricmp(middleClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			middleClickCommand = ("!execute [\"" + iconCommand + "\"]");
		else if (stricmp(middleClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			middleClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else if (stricmp(middleClickCommand.c_str(), ".context") != NULL && stricmp(middleClickCommand.c_str(), ".properties") != NULL)
			middleClickCommand = "";
		
		if (stricmp(middleDoubleClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			middleDoubleClickCommand = ("!execute[\"" + iconCommand + "\"]");
		else if (stricmp(middleDoubleClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			middleDoubleClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else
			middleDoubleClickCommand = "";

		if (stricmp(rightClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			rightClickCommand = ("!execute [\"" + iconCommand + "\"]");
		else if (stricmp(rightClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			rightClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else if (stricmp(rightClickCommand.c_str(), ".context") != NULL && stricmp(rightClickCommand.c_str(), ".properties") != NULL)
			rightClickCommand = "";
		
		if (stricmp(rightDoubleClickCommand.c_str(), ".open") == NULL && !itemIsFolder)
			rightDoubleClickCommand = ("!execute [\"" + iconCommand + "\"]");
		else if (stricmp(rightDoubleClickCommand.c_str(), ".open") == NULL && itemIsFolder)
			rightDoubleClickCommand = ("!execute [explorer.exe /e,\"" + iconCommand + "\"]");
		else
			rightDoubleClickCommand = "";

		//Initial Repositioning
		reposition(actualx, actualy , actualwidth, actualheight);
		
		//Tooltip
		addHint(hWnd, tooltip);

		//Update Region (Magic Pink)
		updateRegion();

		if (hWnd)
		{
			//File Actions
			if (itemIsFolder)
				DragAcceptFiles(hWnd, true);
			else
				DragAcceptFiles(hWnd, false);

			ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
			SetParent(hWnd, GetLitestepDesktop());
			ModifyStyle(hWnd, WS_CHILD, WS_POPUP);
			
			SetWindowPos(hWnd, HWND_NOTOPMOST,
				0, 0, 0, 0,
				SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
			
			//Initial Show
			GetxRCBool("xDesktop", "StartHidden") ? hide() : show();
		}
}

//------------------------------------------------------------
//Reposition
//------------------------------------------------------------
void DesktopItem::reposition(int newx, int newy, int newwidth, int newheight)
{
	if(hWnd)
	{
		actualx = newx;
		actualy = newy;
		//actualheight = newheight;
		//actualwidth = newwidth;
		if (newheight == 0 && newwidth == 0)
		{
			SetWindowPos(hWnd, 0, newx, newy, 0, 0,
				SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
		}
		else
		{
			SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
				SWP_NOACTIVATE | SWP_NOZORDER);
		}
	}
}

//------------------------------------------------------------
//Show/Hide
//------------------------------------------------------------
void DesktopItem::show()
{
	if(hWnd)
	{
		ShowWindow(hWnd, SW_SHOWNOACTIVATE);

		visible = true;
	}
}

void DesktopItem::hide()
{
	if(hWnd)
	{
		ShowWindow(hWnd, SW_HIDE);
		
		visible = false;
	}
}

//------------------------------------------------------------
//Mouse Related Actions
//------------------------------------------------------------

void DesktopItem::onLButtonDblClk(int x, int y)
{
	if(!leftDoubleClickCommand.empty())
		LSExecute(hWnd, leftDoubleClickCommand.c_str(), SW_SHOWNORMAL);
}

void DesktopItem::onLButtonDown(int x, int y)
{
	// 1 = LeftButton
	ontemplateButtonDown(1);
}

void DesktopItem::onLButtonUp(int x, int y)
{	
	// 1 = LeftButton
	ontemplateButtonUp(x, y, 1, leftClickCommand);
}

void DesktopItem::onMButtonDblClk(int x, int y)
{
	if(!middleDoubleClickCommand.empty())
		LSExecute(hWnd, middleDoubleClickCommand.c_str(), SW_SHOWNORMAL);
}

void DesktopItem::onMButtonDown(int x, int y)
{
	// 3 = MiddleButton
	ontemplateButtonDown(3);
}

void DesktopItem::onMButtonUp(int x, int y)
{
	// 3 = MiddleButton
	ontemplateButtonUp(x, y, 3, middleClickCommand);
}

void DesktopItem::onRButtonDblClk(int x, int y)
{
	if(!rightDoubleClickCommand.empty())
		LSExecute(hWnd, rightDoubleClickCommand.c_str(), SW_SHOWNORMAL);
}

void DesktopItem::onRButtonDown(int x, int y)
{
	// 2 = RightButton
	ontemplateButtonDown(2);
}

void DesktopItem::onRButtonUp(int x, int y)
{
	// 2 = RightButton
	ontemplateButtonUp(x, y, 2, rightClickCommand);
}

void DesktopItem::ontemplateButtonDown(int actualMouseClick)
{
	if(moveable && moveButton == actualMouseClick)
	{
		if (!mousePressed)
		{
			mousePressed = true;
			GetCursorPos(&savedpt);
			ScreenToClient(hWnd, &savedpt);
			SetCapture(hWnd);
		}
	}
}

void DesktopItem::ontemplateButtonUp(const int &x, const int &y, int actualMouseClick, const string &clickCommand)
{
	if(moveable && moveButton == actualMouseClick)
	{	
		//To be on the save side
		ReleaseCapture();

		if (mousePressed)
		{
			string bang;
			bang = ("!xdesktopsave " + iconCommand);
			LSExecute(hWnd, bang.c_str(), SW_SHOWNORMAL);	
		}
		mousePressed = false;
	}
	else 
	{
		//To be on the save side
		ReleaseCapture();

		if (!clickCommand.empty() && stricmp(clickCommand.c_str(), ".context") != NULL && stricmp(clickCommand.c_str(), ".properties") != NULL)
			LSExecute(hWnd, clickCommand.c_str(), SW_SHOWNORMAL);
		else if (stricmp(clickCommand.c_str(), ".context") == NULL)
			ContextMenu(iconCommand.c_str());
		else if (stricmp(clickCommand.c_str(), ".properties") == NULL)
		{
			SHELLEXECUTEINFO sei = {sizeof(sei)};
			sei.fMask = SEE_MASK_INVOKEIDLIST;
			sei.lpVerb = "properties";
			sei.lpFile = iconCommand.c_str();
			ShellExecuteEx(&sei);
		}

		mousePressed = false;
	}
}

void DesktopItem::onMouseMove(int x, int y)
{
	if(!mouseInside)
	{
		mouseInside = true;

		if (easyDesktopFocus)
			SendMessage(hWnd, WM_ACTIVATE, WA_CLICKACTIVE, 0);

		SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);
	}

	if (moveable && mousePressed)
	{
		KillTimer(hWnd, TIMER_MOUSETRACK);

		POINT newpt = {x,y};
		ClientToScreen(hWnd, &newpt);
		newpt.x = newpt.x-savedpt.x;
		newpt.y = newpt.y-savedpt.y;

		reposition(newpt.x, newpt.y, 0, 0);
	}
	else
	{
		SetTimer(hWnd, TIMER_MOUSETRACK, 100, 0);
	}
}

void DesktopItem::repaint(bool invalidateCache)
{
	if(hWnd)
	{
		if(invalidateCache)
		{
			if(backgroundDC && backgroundBitmap)
			{
				backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
				DeleteObject(backgroundBitmap);
				backgroundBitmap = 0;
			}
		}

		InvalidateRect(hWnd, NULL, false);
	}
}

void DesktopItem::onPaint(HDC hDC)
{
	int width = actualwidth;
	int height = actualheight;

	// keep a cached rendition of the background
	if(!backgroundBitmap)
	{
		RECT mainRect;
		GetWindowRect(hWnd, &mainRect);

		width = mainRect.right - mainRect.left;
		height = mainRect.bottom - mainRect.top;

		RECT completeFillRect;
		completeFillRect.left = 0;
		completeFillRect.top = 0;
		completeFillRect.right = width;
		completeFillRect.bottom = height;

		if(!backgroundDC)
			backgroundDC = CreateCompatibleDC(hDC);

		backgroundBitmap = CreateCompatibleBitmap(hDC, width, height);
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
		
		//Normal Background Painting
		if ( defaultbackground )
			defaultbackground->apply(backgroundDC, 0, 0, width, height, iconCommand, name, iconOverlay);

		//--------------------------------------------------------------------------------
		// start rendering text
		//--------------------------------------------------------------------------------

		if ( defaultfont )
		{
			defaultfont->apply(backgroundDC,
						leftBorder,
						topBorder,
						width - leftBorder - rightBorder,
						height - topBorder - bottomBorder,
						text,
						align | DT_VCENTER | DT_NOPREFIX | DT_EXPANDTABS | DT_WORDBREAK,
						vertAlign);
		}

		//--------------------------------------------------------------------------------
		// stop rendering text
		//--------------------------------------------------------------------------------
	}

	// blt the double buffer to the display
	BitBlt(hDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);
}

void DesktopItem::updateRegion()
{
	HDC tmpDC = CreateCompatibleDC(backgroundDC);
	HBITMAP tmpBmp = CreateBitmap(actualwidth, actualheight, 1, 32, NULL);
	HGDIOBJ tmpObj = SelectObject(tmpDC, tmpBmp);

	//Draw Image
	if (backgroundDC && backgroundBitmap)
		BitBlt(tmpDC, 0, 0, actualwidth, actualheight, backgroundDC, 0, 0, SRCCOPY);
	else
		onPaint(tmpDC);
			
	SelectObject(tmpDC, tmpObj);

	HRGN region = BitmapToRegion(tmpBmp, RGB(255, 0, 255), 0, 0, 0);
	HRGN windowRgn = CreateRectRgn(0, 0, 0, 0);
	CombineRgn(windowRgn, region, NULL, RGN_COPY);

	if (!SetWindowRgn(hWnd, windowRgn, true))
		DeleteObject(windowRgn);	

	DeleteObject(region);
	DeleteObject(tmpBmp);
	DeleteObject(tmpObj);
	DeleteDC(tmpDC);
}

void DesktopItem::onWindowPosChanging(WINDOWPOS *windowPos)
{
	if (mousePressed && DIsnapto)
	{
		POINT newpt;
		GetCursorPos(&newpt);
			
		POINT pos = {windowPos->x, windowPos->y};
		POINT size = {windowPos->cx, windowPos->cy};

		int result = 0;

		for( int i = 0; i != (GetSystemMetrics(SM_CXSCREEN)-2*DIx)/DIwidth; i++)
		{
			for (int j = 0; j != (GetSystemMetrics(SM_CYSCREEN)-2*DIy)/DIheight; j++)
			{
				result = isCloseTo( pos, size, DIx+i*DIwidth, DIy+j*DIheight, 0, 0, DIsnapdistance);
						
				if (result != 0)
					break;
			}
			if (result != 0)
				break;
		}
				
		if (result == 1)
		{
			windowPos->x = pos.x;
			windowPos->y = newpt.y;
		}
		else if (result == 2)
		{
			windowPos->x = newpt.x;
			windowPos->y = pos.y;
		}
		else if (result == 3)
		{
			windowPos->x = pos.x;
			windowPos->y = pos.y;
		}
	}
}

void DesktopItem::onWindowPosChanged(WINDOWPOS *windowPos)
{
	repaint(true);
}

void DesktopItem::onSize(int width, int height)
{
	actualheight = height;
	actualwidth = width;

	repaint(true);
}

void DesktopItem::onTimer(int timerID)
{
	switch(timerID)
	{
		case TIMER_MOUSETRACK:

			POINT pt;
			RECT rc;
			GetCursorPos(&pt);
			GetWindowRect(hWnd, &rc);

			if(!PtInRect(&rc, pt))
			{
				//To be on the save side
				ReleaseCapture();

				KillTimer(hWnd, TIMER_MOUSETRACK);
				mouseInside = false;
				mousePressed = false;

				if (easyDesktopFocus)
					SendMessage(hWnd, WM_ACTIVATE, WA_INACTIVE, 0);
			}

		break;
	}
}

bool DesktopItem::onWindowMessage(UINT message, WPARAM wParam, LPARAM lParam, LRESULT &lResult)
{
	extern HWND messageHandler;

	switch(message)
	{
		case WM_ACTIVATE:
		{
			if (wParam == WA_INACTIVE)
			{
				iconOverlay = false;
				repaint(true);
			}
			else if (wParam == WA_CLICKACTIVE)
			{
				iconOverlay = true;
				repaint(true);
				SendMessage(hWnd, WM_PAINT, 0, 0);
			}

			return true;
		}

		case WM_CLOSE:
		{
			lResult = 0;
			return true;
		}

		case WM_DROPFILES:
		{
			if(dropCommand.empty())	// we shouldn't get here... but somehow we did :p
				break;

			int intFileAction;

			char action[32], szTarget[MAX_PATH];;
			char *buffers[] = {action, szTarget};

			LCTokenize(dropCommand.c_str(), buffers, 2, NULL);

			if (stricmp(action, "move") == 0)
				intFileAction = 1;
			else if (stricmp(action, "copy") == 0)
				intFileAction = 2;
			else if (stricmp(action, "delete") == 0)
				intFileAction = 0;

			char *szSourceFiles = NULL;
			char buffer[MAX_PATH];
			int numFiles;
			UINT len,scursor = 0;

			numFiles = DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, NULL);

			szSourceFiles = new char[MAX_PATH*numFiles+1];
			scursor = 0;

			for (int i=0;i<numFiles;i++)
			{
				len = DragQueryFile((HDROP) wParam, i, buffer, MAX_PATH);
					
				strcpy(szSourceFiles+scursor,buffer);
				scursor += len + 1;
			}
					
			DragFinish((HDROP) wParam);

			string check = szTarget;
			if (check.substr(1, check.length()-1) == "\\")
				check.append( "*" );
			else
				check.append( "\\*" );
				
			//Test Path for Existence
			WIN32_FIND_DATA FindData;
			HANDLE hFindHandle = FindFirstFile(check.c_str(), &FindData);

			if(hFindHandle != INVALID_HANDLE_VALUE || intFileAction == 0)
			{
				szSourceFiles[scursor] = 0;
				strcpy(buffer, szTarget);
				buffer[strlen(buffer)+1] = 0;
					
				SHFILEOPSTRUCT shfop;
				shfop.hwnd = NULL;
				if (intFileAction == 1)
					shfop.wFunc = FO_MOVE;
				else if (intFileAction == 2)
					shfop.wFunc = FO_COPY;
				else if (intFileAction == 0)
					shfop.wFunc = FO_DELETE;
				shfop.pFrom = szSourceFiles;
				shfop.pTo = buffer;
				shfop.fFlags = FOF_NOCONFIRMMKDIR | FOF_ALLOWUNDO;
				SHFileOperation(&shfop);
			}
				
			delete[] szSourceFiles;
			
			return 0;
		}
		
		case WM_LBUTTONDBLCLK:
		{
			onLButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_LBUTTONDOWN:
		{
			onLButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_LBUTTONUP:
		{
			onLButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONDBLCLK:
		{
			onMButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONDOWN:
		{
			onMButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONUP:
		{
			onMButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONDBLCLK:
		{
			onRButtonDblClk((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONDOWN:
		{
			onRButtonDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONUP:
		{
			onRButtonUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_MOUSEMOVE:
		{
			onMouseMove((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hDC;

			if(!wParam)
				hDC = BeginPaint(hWnd, &ps);
			else
				hDC = (HDC) wParam;

			onPaint(hDC);

			if(!wParam)
				EndPaint(hWnd, &ps);

			return true;
		}

		case WM_SIZE:
		{
			onSize((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
			return true;
		}

		case WM_TIMER:
		{
			onTimer((int) wParam);
			return true;
		}

		case WM_WINDOWPOSCHANGING:
		{
			onWindowPosChanging((WINDOWPOS *) lParam);
			break;
		}

		case WM_WINDOWPOSCHANGED:
		{
			onWindowPosChanged((WINDOWPOS *) lParam);
			return false;
		}
	}

	return false;
}

LRESULT DesktopItem::windowProcedure(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	DesktopItem *desktopitem = NULL;

	if(message == WM_NCCREATE)
	{
		desktopitem = (DesktopItem *) ((CREATESTRUCT *) lParam)->lpCreateParams;
		desktopitem->hWnd = hWnd;
		SetWindowLong(hWnd, 0, (LONG) desktopitem);
	}
	else
		desktopitem = (DesktopItem *) GetWindowLong(hWnd, 0);

	if(desktopitem)
	{
		LRESULT lResult = 0;

		if(desktopitem->onWindowMessage(message, wParam, lParam, lResult))
			return lResult;
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

//------------------------------------------------------------
//Tooltips (Add, Remove)
//------------------------------------------------------------

void DesktopItem::addHint(HWND hWnd, LPSTR caption)
{
	TOOLINFO ti;
	RECT clientRect;

	if (!TooltipHints)
		return ;
	GetClientRect(hWnd, &clientRect);
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_SUBCLASS;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = clientRect;
	ti.hinst = hInstance;
	ti.lpszText = caption;
	ti.lParam = 0;
	SendMessage(TooltipHints, TTM_ADDTOOL, 0, LPARAM(&ti));
}

void DesktopItem::removeHint(HWND hWnd)
{
	TOOLINFO ti;
	RECT emptyRect = {0, 0, 0, 0};

	if (!TooltipHints)
		return ;
	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = 0;
	ti.hwnd = hWnd;
	ti.uId = 0;
	ti.rect = emptyRect;
	ti.hinst = hInstance;
	ti.lpszText = NULL;
	ti.lParam = 0;
	SendMessage(TooltipHints, TTM_DELTOOL, 0, LPARAM(&ti));
}

int DesktopItem::isCloseTo(POINT &pos, const POINT size, const int x, const int y, const int cx, const int cy, const int span)
{
	bool internalX = false;
	bool internalY = false;

	if ((pos.x > x-span) && (pos.x < x+span))
	{
		if ( ((pos.y > y-span) && (pos.y < y+span)) 
			|| ((pos.y+size.y > y-span) && (pos.y+size.y < y+span)) 
			|| ((pos.y > y+cy-span ) && (pos.y < y+cy+span)) 
			|| ((pos.y+size.y > y+cy-span ) && (pos.y+size.y < y+cy+span)) )
		{
			pos.x = x;
			internalX = true;
		}
	}

	if ((pos.y > y-span) && (pos.y < y+span))
	{
		if ( ((pos.x > x-span) && (pos.x < x+span)) 
			|| ((pos.x+size.x > x-span) && (pos.x+size.x < x+span)) 
			|| ((pos.x > x+cx-span ) && (pos.x < x+cx+span)) 
			|| ((pos.x+size.x > x+cx-span ) && (pos.x+size.x < x+cx+span)) )
		{
			pos.y = y;
			internalY = true;
		}
	}

	if (!internalX && !internalY)
		return 0;
	else if (internalX && !internalY)
		return 1;
	else if (!internalX && internalY)
		return 2;
	else
		return 3;
}