#include "stdafx.h"
#include "DesktopItem.h"
#include "DesktopTexture.h"

extern DesktopItemList desktopitemList;

DesktopTexture::DesktopTexture()
{
	useSmallIcon = false;

	noActiveIconColor = false;
}

DesktopTexture::~DesktopTexture()
{
}

void DesktopTexture::configure()
{
	NameValuePair iconTextValues[] = {
		{ "top", 1 },
		{ "bottom", 2 },
		{ "left", 3 },
		{ "right", 4 },
		{ 0, 0 }
	};

	noActiveIconColor = GetxRCBool("xdesktop" , "NoActiveIconColor");

	normal_saturation = GetxRCInt("xDesktop" , "SaturationIntensity", 255, 0, 255);
	normal_hueIntensity = GetxRCInt("xDesktop" , "HueIntensity", 0, 0, 255);
	normal_hueColor = GetxRCColor("xDesktop", "HueColor", RGB(0, 0, 0));

	active_saturation = GetxRCInt("xDesktop" , "ActiveSaturationIntensity", normal_saturation, 0, 255);
	active_hueIntensity = GetxRCInt("xDesktop" , "ActiveHueIntensity", normal_hueIntensity, 0, 255);
	active_hueColor = GetxRCColor("xDesktop", "ActiveHueColor", normal_hueColor);

	iconSize = GetxRCInt("xdesktop" , "IconSize", 32, 16, 64);

	iconTextPosition = GetxRCNamedValue("xdesktop", "TextPosition", iconTextValues, 2);
	iconTextSpacing = GetxRCInt("xdesktop" , "TextSpacing", 5);

	useSmallIcon = GetxRCBool("xdesktop" , "UseSmallIcon");
	
	int color;
	char main[32], light[32], dark[32];
	char *buffers[] = {main, light, dark};

	int numTokens = LCTokenize((GetxRCLine("xdesktop", "TextBackground", "ff00ff")).c_str(), buffers, 3, NULL);

	if (numTokens == 3)
	{
		color = strtol(light, NULL, 16);
		iconTextBackgroundLight = RGB(GetBValue(color), GetGValue(color), GetRValue(color));

		color = strtol(dark, NULL, 16);
		iconTextBackgroundDark = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
	}
	else
	{
		iconTextBackgroundLight = RGB(255,0,255);
		iconTextBackgroundDark = RGB(255,0,255);
	}
	if(numTokens >= 1)
	{
		color = strtol(main, NULL, 16);
		iconTextBackground = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
	}
	if (numTokens == 2)
	{
		color = strtol(light, NULL, 16);
		iconTextBackgroundLight = RGB(GetBValue(color), GetGValue(color), GetRValue(color));
		iconTextBackgroundDark = iconTextBackgroundLight;
	}
}

void DesktopTexture::apply(HDC hDC, int x, int y, int width, int height, const string &iconCommand, const string &name, bool activated)
{
	HICON hIcon = NULL;

	SHFILEINFO shFileInfo;
    SHGetFileInfo(iconCommand.c_str(), NULL, &shFileInfo, sizeof(shFileInfo), SHGFI_ICON | ( (activated && !noActiveIconColor) ? SHGFI_SELECTED : 0 ) | (useSmallIcon ? SHGFI_SMALLICON : SHGFI_LARGEICON));
	hIcon = shFileInfo.hIcon;
	
	if(hIcon == NULL)
		return;

	HDC	memoryDC = CreateCompatibleDC(hDC);
	HBITMAP memoryBitmap = CreateCompatibleBitmap(hDC, width, height);
	SelectObject(memoryDC, memoryBitmap);

	RECT r;
	r.top = 0;
	r.left = 0;
	r.right = width;
	r.bottom = height;

	::SetBkColor(memoryDC, RGB(255, 0, 255));
	::ExtTextOut(memoryDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

	int xpos, ypos;
	if (iconTextPosition == 1)
	{
		xpos = (width-iconSize)/2;
		ypos = height-iconSize;
		r.top = 0;
		r.left = 0;
		r.right = width;
		r.bottom = height-iconSize-iconTextSpacing;
	}
	else if (iconTextPosition == 2)
	{
		xpos = (width-iconSize)/2;
		ypos = 0;
		r.top = iconSize+iconTextSpacing;
		r.left = 0;
		r.right = width;
		r.bottom = height;
	}
	else if (iconTextPosition == 3)
	{
		xpos = width-iconSize;
		ypos = (height-iconSize)/2;
		r.top = 0;
		r.left = 0;
		r.right = width-iconSize-iconTextSpacing;
		r.bottom = height;
	}
	else if (iconTextPosition == 4)
	{
		xpos = 0;
		ypos = (height-iconSize)/2;
		r.top = 0;
		r.left = iconSize+iconTextSpacing;
		r.right = width;
		r.bottom = height;
	}

	::SetBkColor(memoryDC, iconTextBackgroundDark);
	::ExtTextOut(memoryDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

	r.right = r.right-1;
	r.bottom = r.bottom-1;
	::SetBkColor(memoryDC, iconTextBackgroundLight);
	::ExtTextOut(memoryDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

	r.left = r.left+1;
	r.top = r.top+1;
	::SetBkColor(memoryDC, iconTextBackground);
	::ExtTextOut(memoryDC, 0, 0, ETO_OPAQUE, &r, NULL, 0, NULL);

	int curx = 0;
	int cury = 0;
	for ( DesktopItemListIterator iter = desktopitemList.begin(); iter != desktopitemList.end(); iter++)
	{
		if(stricmp((*iter)->getName().c_str(), name.c_str()) == NULL)
		{
			curx = (*iter)->getX();
			cury = (*iter)->getY();
			break;
		}
	}

	if (hIcon)
	{
		if ( (normal_saturation != 255 || normal_hueIntensity != 0) || (activated && (active_saturation != 255 || active_hueIntensity != 0)) )
		{
			//Normal
			HDC memoryColorIconDC = CreateCompatibleDC(0);
			HBITMAP memoryColorIconBitmap = CreateBitmap(iconSize, iconSize, 1, 32, NULL);
			SelectObject(memoryColorIconDC, memoryColorIconBitmap);
		
			DrawIconEx(memoryColorIconDC, 0, 0, hIcon, iconSize, iconSize, 0, NULL, DI_NORMAL);

			//Mask
			HDC memoryMaskIconDC = CreateCompatibleDC(memoryColorIconDC);
			HBITMAP memoryMaskIconBitmap = CreateCompatibleBitmap(memoryMaskIconDC, iconSize, iconSize);
			SelectObject(memoryMaskIconDC, memoryMaskIconBitmap);

			DrawIconEx(memoryMaskIconDC, 0, 0, hIcon, iconSize, iconSize, 0, NULL, DI_MASK);

			int _saturation = (activated ? active_saturation : normal_saturation);
			int _hueIntensity = (activated ? active_hueIntensity : normal_hueIntensity);
			COLORREF _hueColor = (activated ? active_hueColor : normal_hueColor);

			HDC	memoryPaintDC = CreateCompatibleDC(hDC);
			HBITMAP memoryPaintBitmap = CreateCompatibleBitmap(hDC, iconSize, iconSize);
			SelectObject(memoryPaintDC, memoryPaintBitmap);

			if ( IsOS(OS_XP) )
				PaintDesktopEx(memoryPaintDC, 0, 0, iconSize, iconSize, curx+xpos, cury+ypos, false);

			int iy = 0;
			int ix = 0;
			COLORREF cl = 0;
			BYTE r = 0;
			BYTE g = 0;
			BYTE b = 0;
			for(iy = 0; iy < iconSize; iy++)
			{
				for(ix = 0; ix < iconSize; ix++)
				{
					// Loop through all transparent pixels...
					// If the mask-pixel is white, then break
					if(!GetPixel(memoryMaskIconDC, ix, iy))
					{
						cl = GetPixel(memoryColorIconDC, ix, iy);

						r = GetRValue(cl);
						g = GetGValue(cl);
						b = GetBValue(cl);

						// Saturation effect
						if(_saturation != 255)
						{
							BYTE gray = (BYTE)(r * 0.3086 + g * 0.6094 + b * 0.0820);

							r = (BYTE)((r * _saturation + gray * (255 - _saturation) + 255) >> 8);
							g = (BYTE)((g * _saturation + gray * (255 - _saturation) + 255) >> 8);
							b = (BYTE)((b * _saturation + gray * (255 - _saturation) + 255) >> 8);
						}

						// Hue effect
						if(_hueIntensity != 0)
						{
							// The B & R values of the hue is swapped here, can somebody tell me why it only works that way?
							// test normal
							r = (BYTE)((r * (255 - _hueIntensity) + GetRValue(_hueColor) * _hueIntensity + 255) >> 8);
							g = (BYTE)((g * (255 - _hueIntensity) + GetGValue(_hueColor) * _hueIntensity + 255) >> 8);
							b = (BYTE)((b * (255 - _hueIntensity) + GetBValue(_hueColor) * _hueIntensity + 255) >> 8);
						}

						SetPixelV(memoryPaintDC, ix, iy, RGB(r, g, b));
					}
				}
			}

			BitBlt( memoryDC, xpos, ypos, iconSize, iconSize, memoryMaskIconDC, 0, 0, SRCAND );
			// Combine the foreground with the background
			BitBlt( memoryDC, xpos, ypos, iconSize, iconSize, memoryPaintDC, 0, 0, (IsOS(OS_XP) ? SRCCOPY : SRCPAINT));

			memoryPaintBitmap = (HBITMAP) SelectObject(memoryPaintDC, memoryPaintBitmap);
			DeleteObject(memoryPaintBitmap);
			DeleteDC(memoryPaintDC);

			memoryColorIconBitmap = (HBITMAP) SelectObject(memoryColorIconDC, memoryColorIconBitmap);
			DeleteObject(memoryColorIconBitmap);
			DeleteDC(memoryColorIconDC);

			memoryMaskIconBitmap = (HBITMAP) SelectObject(memoryMaskIconDC, memoryMaskIconBitmap);
			DeleteObject(memoryMaskIconBitmap);
			DeleteDC(memoryMaskIconDC);
		}
		else
		{
			if ( IsOS(OS_XP) )
				PaintDesktopEx(memoryDC, xpos, ypos, iconSize, iconSize, curx+xpos, cury+ypos, false);
			DrawIconEx(memoryDC, xpos, ypos, hIcon, iconSize, iconSize, 0, NULL, DI_NORMAL);
		}
	}

	BitBlt( hDC, x, y, width, height, memoryDC, 0, 0, SRCCOPY );

	memoryBitmap = (HBITMAP) SelectObject(memoryDC, memoryBitmap);
	DeleteObject(memoryBitmap);
	DeleteDC(memoryDC);

	if ( hIcon )
		DestroyIcon(hIcon);
}


