#include "StdAfx.h"
#include "DesktopItem.h"
#include "DesktopTexture.h"
#include "Font.h"

#include "FileInformation.h"
#include "NotifyDirCheck.h"

#include "regexp.h"
#include "regmagic.h"

void getToken(char line[4096], char temp[4096], bool escape_codes = true);

void TextAppend(HWND hwndCaller, LPCSTR pszArgs);
void TextDelete(HWND hwndCaller, LPCSTR pszArgs);
void TextReplace(HWND hwndCaller, LPCSTR pszArgs);

StringList getActualDesktopItemPaths(bool defaultDesktop, string searchPath);
void DesktopUpdate(int updateMode, string updatedFile, StringList ActualDesktopItemPaths);

void DesktopSave(HWND hwndCaller, LPCSTR pszArgs);

void BangDestroy(HWND hwndCaller, LPCSTR pszArgs);
void BangCreate(HWND hwndCaller, LPCSTR pszArgs);
void BangHide(HWND hwndCaller, LPCSTR pszArgs);
void BangShow(HWND hwndCaller, LPCSTR pszArgs);
void BangToggle(HWND hwndCaller, LPCSTR pszArgs);

void OnStart(LPCSTR path);
void OnStop();

bool customDelete = false;

bool useautodesktop = GetxRCBool("xDesktop", "AutoCreate");
string desktopSavePath = GetxRCString("xDesktop", "ConfigFile", "");

class CNotifyDirCheck;

typedef list<CNotifyDirCheck *> DirCheckList;
typedef list<CNotifyDirCheck *>::iterator DirCheckListIterator;

DirCheckList dircheckList;

CStringList pathList;

string Bang = "";

StringList stringDesktopItemList;

DesktopItemList desktopitemList;

Font *defaultfont = NULL;
DesktopTexture *defaultbackground = NULL;

HINSTANCE hInstance;
HWND messageHandler;

int lsMessages[] = {
	LM_GETREVID,
	0
};

#define LM_UPDATEBG (WM_USER + 1)

LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case LM_GETREVID:
		{
			UINT uLength;
			StringCchPrintf((char*)lParam, 64, "%s.dll: %s", V_NAME, V_VERSION);
			
			if (SUCCEEDED(StringCchLength((char*)lParam, 64, &uLength)))
				return uLength;

			lParam = NULL;
			return 0;
		}

		case LM_UPDATEBG:
		{
			PaintDesktopEx(0, 0, 0, 0, 0, 0, 0, TRUE);

			for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
				(*diit)->repaint(true);

			return 0;
		}

		case WM_SETTINGCHANGE:
		{
			PostMessage(hWnd, LM_UPDATEBG, 0, 0);
			return 0;
		}
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes);

// Why are we advoiding the crt startup code? The start up code is not large
// enough to cause bloat (except possibly a long system call, and I suspect that
// would need done eventually anyway) (default _DllMainCRTStartup is in
// (MSVCDIR\crt\src\dllcrt0.c) The elimination of the call to _CRT_INIT is 
// also causing warnings with MSVC7
BOOL __stdcall _DllMainCRTStartup(HINSTANCE hInst, DWORD fdwReason, LPVOID lpvRes)
{
	// We don't need thread notifications for what we're doing.  Thus, get
	// rid of them, thereby eliminating some of the overhead of this DLL
	DisableThreadLibraryCalls(hInst);
	return TRUE;
}

int initModuleEx(HWND hParent, HINSTANCE hInstance, const char *lsPath)
{
	WNDCLASSEX wc;

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS | CS_DBLCLKS;
	wc.lpfnWndProc = DesktopItem::windowProcedure;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = sizeof(DesktopItem *);
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = 0;
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xDesktopItem";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = MessageHandlerProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hbrBackground = 0;
	wc.hCursor = 0;
	wc.hIcon = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "xDesktopItemMessageHandler";
	wc.hIconSm = 0;

	RegisterClassEx(&wc);

	messageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
		"xDesktopItemMessageHandler",
		0,
		WS_POPUP,
		0, 0, 0, 0, 
		0,
		0,
		hInstance,
		0);

	if (!messageHandler)
		return 1;
	
	SendMessage(GetLitestepWnd(),
		LM_REGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	::hInstance = hInstance;

	//Check LoadModule Order
	if ( FindWindow("DesktopBackgroundClass", NULL) == NULL && !GetxRCBool("xDesktop", "NoDesktopWarning"))
		MessageBox(NULL, "Just To Prevent A \"Bug Report\"!\n\nPlease Load A DesktopModule (Desktop2 or jDesk) BEFORE xDesktop!\nOtherwise You Will Get Annoying Z-Order Problems!\n\nBtw.: This Message Will Pop Up On Every Recycle Till You Fix It :P", "Important User Information", MB_SETFOREGROUND | MB_ICONEXCLAMATION );

	//Default Font
	defaultfont = new Font();
	defaultfont->configure("xdesktop");
	//Default Texture
	defaultbackground = new DesktopTexture();
	defaultbackground->configure();

	//Desktop Items
	stringDesktopItemList = GetxRCLineList("*xdesktopitem");

	if (!stringDesktopItemList.empty())
	{	
		for(StringListIterator sit = stringDesktopItemList.begin(); sit != stringDesktopItemList.end(); sit++)
		{
			char name[32];
			char *buffers[] = {name};

			if(LCTokenize((*sit).c_str(), buffers, 1, 0) >= 1)
			{
				DesktopItem *desktopitem = new DesktopItem(name);

				desktopitemList.insert(desktopitemList.end(), desktopitem);

				desktopitem->desktopload(hInstance, (*sit));
			}
		}
	}

	AddBangCommand("!xDesktopSave", DesktopSave);

	AddBangCommand("!xDesktopCreate", BangCreate);
	AddBangCommand("!xDesktopDestroy", BangDestroy);
	AddBangCommand("!xDesktopHide", BangHide);
	AddBangCommand("!xDesktopShow", BangShow);
	AddBangCommand("!xDesktopToggle", BangToggle);

	if (useautodesktop)
	{
		if (stringDesktopItemList.empty())
		{
			Bang = "!xdesktopcreate replace \".desktop\"";
			LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
		}

		char path[MAX_PATH];			
		VarExpansion( path, "$desktopdir$" );
		OnStart( path );
		VarExpansion( path, "$commondesktopdir$" );
		OnStart( path );
	}

	return 0;
}

extern HDC hdcDesktop;
extern HBITMAP hbmDesktop;

void quitModule(HINSTANCE hInstance)
{
	RemoveBangCommand("!xDesktopSave");

	RemoveBangCommand("!xDesktopCreate");
	RemoveBangCommand("!xDesktopDestroy");
	RemoveBangCommand("!xDesktopHide");
	RemoveBangCommand("!xDesktopShow");
	RemoveBangCommand("!xDesktopToggle");

	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
		delete *diit;

	desktopitemList.clear();

	delete defaultfont;

	delete defaultbackground;

	if (useautodesktop || customDelete)
	{
		if (!stringDesktopItemList.empty())
		{	
			int counter = 1;

			//Clear all *xdesktopitem lines
			Bang = ("@" + desktopSavePath + "@ @\\*[xX][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]@");
			TextDelete(NULL, Bang.c_str());

			for(StringListIterator sit = stringDesktopItemList.begin(); sit != stringDesktopItemList.end(); sit++)
			{
				char buffer[32];
				strcpy(buffer, "xldi");
				char counterbuffer[16];
				itoa( counter, counterbuffer, 10 );
				if (counter < 10)
					strcat(buffer, "0");
				strcat(buffer, counterbuffer);

				char name[64], data[MAX_LINE_LENGTH];
				char *linebuffers[] = {name};

				LCTokenize((*sit).c_str(), linebuffers, 1, data);

				Bang = ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + data + "@");
				TextAppend(NULL, Bang.c_str());

				counter++;
			}

			stringDesktopItemList.clear();
		}

		OnStop();
	}

	SendMessage(GetLitestepWnd(),
		LM_UNREGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	DestroyWindow(messageHandler);

	UnregisterClass("xDesktopItem", hInstance);
	UnregisterClass("xDesktopItemMessageHandler", hInstance);

	if (hbmDesktop != 0 || hdcDesktop != 0)
	{
		hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
		DeleteDC(hdcDesktop);
		DeleteObject(hbmDesktop);
	}
}

void RemoveDesktopItem(string path)
{
	bool found = false;
	for(DesktopItemListIterator lit = desktopitemList.begin(); lit != desktopitemList.end(); lit++)
	{
		if (stricmp((*lit)->getIconCommand().c_str(), path.c_str()) == NULL)
		{
			found = true;
			break;
		}
	}
			
	if (found)
	{
		char xname[32], xx[16], xy[16], xpath[MAX_PATH_LENGTH];
		char *buffers[] = {xname, xx, xy, xpath};

		for(StringListIterator sit = stringDesktopItemList.begin(); sit != stringDesktopItemList.end(); sit++)
		{
			LCTokenize((*sit).c_str(), buffers, 4, NULL);

			if (stricmp((*lit)->getIconCommand().c_str(), xpath) == NULL)
			{
				stringDesktopItemList.remove( *sit );
				break;
			}
		}

		(*lit)->hide();
		desktopitemList.remove((*lit));
		delete (*lit);

		if (!useautodesktop)
			customDelete = true;
	}
}

UINT DirCallback( CFileInformation fiObject, EFileAction faAction, LPVOID lpData )
{
	CString Bang;
	CString filepath = fiObject.GetFilePath();

	bool found = false;
	POSITION m_position = pathList.GetHeadPosition();
	while (m_position != NULL && found == FALSE)
	{
		if ( fiObject.IsSomeFileDir( pathList.GetAt(m_position) ) )
			found = TRUE;
		pathList.GetNext(m_position);
	}

	if (found)
	{
		if( IS_CREATE_FILE( faAction ) )
		{
			Bang = ("!xdesktopcreate updateadd \"" + filepath + "\"");
			LSExecute(NULL, Bang, SW_SHOWNORMAL);
		}
		else if( IS_DELETE_FILE( faAction ) )
		{
			Bang = ("!xdesktopcreate updatedel \"" + filepath + "\"");
			LSExecute(NULL, Bang, SW_SHOWNORMAL);
		}
		else if( IS_CHANGE_FILE( faAction ) )
		{
			Bang = ("!xdesktopcreate updateren \"" + filepath + "\"");
			LSExecute(NULL, Bang, SW_SHOWNORMAL);
		}
		else
		{
			return 1; //error, stop thread
		}
	}

	return 0; //success
}

void OnStart(LPCSTR path) 
{
	CNotifyDirCheck *DirCheck = new CNotifyDirCheck();
	
	dircheckList.insert(dircheckList.end(), DirCheck);
		
	DirCheck->SetDirectory( path );
	pathList.AddTail( path );

	//set your callback to work with each new event
	DirCheck->SetActionCallback( DirCallback );

	DirCheck->Run();
}

void OnStop() 
{
	for(DirCheckListIterator dcit = dircheckList.begin(); dcit != dircheckList.end(); dcit++)
	{
		(*dcit)->Stop();
		delete *dcit;
	}

	pathList.RemoveAll();
	dircheckList.clear();
}

//Read all required Files/Folder Paths (in StringList)
StringList getActualDesktopItemPaths(bool defaultDesktop, string searchPath)
{
	StringList ActualDesktopItemPaths;

	string wildcardPath = "";
	string itemName = "";
	
	char tmp[MAX_PATH];
	string tmpHelper = "";

	HANDLE hFindHandle;
	WIN32_FIND_DATA FindData;

	if (defaultDesktop)
	{
		//Create 4 Main ShellFolders: Workspace, My Documents, NetworkHood, Trash 
		if ( !GetxRCBool("xdesktop", "NoWorkSpace") )
			ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{20D04FE0-3AEA-1069-A2D8-08002B30309D}");
		if ( !GetxRCBool("xdesktop", "NoMyDocuments") )
			ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{450D8FBA-AD25-11D0-98A8-0800361B1103}");
		if ( !GetxRCBool("xdesktop", "NoNetwork") )
			ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{208D2C60-3AEA-1069-A2D7-08002B30309D}");
		if ( !GetxRCBool("xdesktop", "NoTrashcan") )
			ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), "::{645FF040-5081-101B-9F08-00AA002F954E}");

		VarExpansion( tmp, "$commondesktopdir$" );
		searchPath = tmp;
		wildcardPath = (searchPath + "*.*");

		hFindHandle = FindFirstFile(wildcardPath.c_str(), &FindData);

		for(;;)
		{
			if(INVALID_HANDLE_VALUE == hFindHandle)
				break;

			itemName = FindData.cFileName;

			//Make sure its a valid folder
			if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
			{
				tmpHelper = (searchPath + itemName);
				ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), tmpHelper.c_str());
			}

			if(!FindNextFile(hFindHandle, &FindData))
				break;
		}

		VarExpansion( tmp, "$desktopdir$" );
		searchPath = tmp;
		wildcardPath = (searchPath + "*.*");

		hFindHandle = FindFirstFile(wildcardPath.c_str(), &FindData);

		for(;;)
		{
			if(INVALID_HANDLE_VALUE == hFindHandle)
				break;
					
			itemName = FindData.cFileName;

			//Make sure its a valid folder
			if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
			{
				tmpHelper = (searchPath + itemName);
				ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), tmpHelper.c_str());
			}

			if(!FindNextFile(hFindHandle, &FindData))
				break;
		}
	}
	else
	{
		//Check for ending BackSlash
		tmpHelper = searchPath.substr(searchPath.length()-1, searchPath.length());

		if (tmpHelper != "\\")
			wildcardPath = (searchPath + "\\*.*");
		else
			wildcardPath = (searchPath + "*.*");

		hFindHandle = FindFirstFile(wildcardPath.c_str(), &FindData);

		for(;;)
		{
			if(INVALID_HANDLE_VALUE == hFindHandle)
				break;
					
			itemName = FindData.cFileName;

			//Make sure its a valid folder
			if (stricmp(itemName.c_str(), ".") != NULL && stricmp(itemName.c_str(), "..") != NULL)
			{
				tmpHelper = (searchPath + itemName);
				ActualDesktopItemPaths.insert(ActualDesktopItemPaths.end(), tmpHelper.c_str());
			}

			if(!FindNextFile(hFindHandle, &FindData))
				break;
		}
	}

	FindClose(hFindHandle);

	return ActualDesktopItemPaths;
}

//Write all Items to File
void DesktopUpdate(int updateMode, string updatedFile, StringList ActualDesktopItemPaths)
{
	int counter = 1;

	//Saves already present Items, Only if:
	//	1. They exist
	//  2. It's wished
	//  3. AutoCreateDesktop isn't running
	if (!stringDesktopItemList.empty() && updateMode == 1 && !useautodesktop)
	{
		for(StringListIterator sit = stringDesktopItemList.begin(); sit != stringDesktopItemList.end(); sit++)
		{
			char buffer[32];
			strcpy(buffer, "xldi");
			char counterbuffer[16];
			itoa( counter, counterbuffer, 10 );
			if (counter < 10)
				strcat(buffer, "0");
			strcat(buffer, counterbuffer);

			char name[64], data[MAX_LINE_LENGTH];
			char *linebuffers[] = {name};

			LCTokenize((*sit).c_str(), linebuffers, 1, data);

			Bang = ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + data + "@");
			TextAppend(NULL, Bang.c_str());

			counter++;
		}
	}
	
	//Saves new Items, Only if:
	//	1. It's wished (Replace)
	//  2. It's wished (Append)
	//  3. None exists (Initial Creation required)
	//  AND IT'S NOT AN UPDATE
	if ( (updateMode == 0 || updateMode == 1 || stringDesktopItemList.empty()) && updateMode != -1)
	{
		int tempcounter = 1;

		NameValuePair directionValues[] = {
		{ "left", 1 },
		{ "right", 2 },
		{ "up", 3 },
		{ "down", 4 },
		{ 0, 0 }
		};

		NameValuePair wrapdirectionValues[] = {
		{ "up", 1 },
		{ "down", 2 },
		{ "left", 3 },
		{ "right", 4 },
		{ 0, 0 }
		};

		int DIwidth = GetxRCInt("xdesktop", "ItemWidth", 100);
		int DIheight = GetxRCInt("xdesktop", "ItemHeight", 75);
		int xSpace = GetxRCInt("xdesktop", "SpacingX", 0);
		int ySpace = GetxRCInt("xdesktop", "SpacingY", 0);
				
		int direction = GetxRCNamedValue("xdesktop", "Direction", directionValues, 2);
		int wrapDirection = GetxRCNamedValue("xdesktop", "WrapDirection", wrapdirectionValues, 2);
		
		int wrapCount;
		if (wrapDirection == 1 || wrapDirection == 2)
			wrapCount = GetxRCInt("xdesktop", "WrapCount", (GetSystemMetrics(SM_CXSCREEN)-xSpace)/(DIwidth+xSpace));
		else if (wrapDirection == 3 || wrapDirection == 4)
			wrapCount = GetxRCInt("xdesktop", "WrapCount", (GetSystemMetrics(SM_CYSCREEN)-ySpace)/(DIheight+ySpace));

		int DIStartx = GetxRCInt("xdesktop", "X", ((GetSystemMetrics(SM_CXSCREEN)-(wrapCount*(DIwidth+xSpace)-xSpace))/2 < 0 ? 0 : (GetSystemMetrics(SM_CXSCREEN)-(wrapCount*(DIwidth+xSpace)-xSpace))/2), 0, GetSystemMetrics(SM_CXSCREEN));
		int DIStarty = GetxRCInt("xdesktop", "Y", ((GetSystemMetrics(SM_CYSCREEN)-(wrapCount*(DIheight+ySpace)-ySpace))/2 < 50 ? 50 : (GetSystemMetrics(SM_CYSCREEN)-(wrapCount*(DIheight+ySpace)-ySpace))/2), 0, GetSystemMetrics(SM_CYSCREEN));

		int tempx;
		int tempy;

		if (direction == 1)
			tempx = DIStartx+DIwidth+xSpace;
		else if (direction == 2)
			tempx = DIStartx-(DIwidth+xSpace);
		else if (direction == 3)
			tempy = DIStarty+DIheight+ySpace;
		else if (direction == 4)
			tempy = DIStarty-(DIheight+ySpace);

		if (direction == 1 || direction == 2)
			tempy = DIStarty;
		else if (direction == 3 || direction == 4)
			tempx = DIStartx;

		for(StringListIterator it = ActualDesktopItemPaths.begin(); it != ActualDesktopItemPaths.end(); it++)
		{
			char buffer[32];
			strcpy(buffer, "xldi");
			char counterbuffer[16];
			char xbuffer[16];
			char ybuffer[16];
			itoa( counter, counterbuffer, 10 );
			if (counter < 10)
				strcat(buffer, "0");
			strcat(buffer, counterbuffer);

			//Sorting/Positioning due to settings
			div_t div_result = div( tempcounter, wrapCount );

			if (direction == 1 || direction == 2)
			{
				if (wrapDirection > 2)
					wrapDirection = 2;
						
				if (div_result.quot > 0 && div_result.rem == 1)
				{
					tempx = DIStartx;
					if (wrapDirection == 1)
						tempy = DIStarty-(div_result.quot*(DIheight+ySpace));
					else if (wrapDirection == 2)
						tempy = DIStarty+div_result.quot*(DIheight+ySpace);
				}
				else
				{
					if (direction == 1)
						tempx = tempx-(DIwidth+xSpace);
					else if (direction == 2)
						tempx = tempx+DIwidth+xSpace;
				}
			}
			if (direction == 3 || direction == 4)
			{
				if (wrapDirection < 3)
					wrapDirection = 4;
					
				if (div_result.quot > 0 && div_result.rem == 1)
				{
					tempy = DIStarty;
					if (wrapDirection == 3)
						tempx = DIStartx-(div_result.quot*(DIwidth+xSpace));
					else if (wrapDirection == 4)
						tempx = DIStartx+div_result.quot*(DIwidth+xSpace);
				}
				else
				{
					if (direction == 3)
						tempy = tempy-(DIheight+ySpace);
					else if (direction == 4)
						tempy = tempy+DIheight+ySpace;
				}
			}

			itoa( tempx, xbuffer, 10 );
			itoa( tempy, ybuffer, 10 );

			if (!useautodesktop)
			{
				Bang = ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + xbuffer + " " + ybuffer + " \"" + (*it) + "\"@");
				TextAppend(NULL, Bang.c_str());
			}

			DesktopItem *desktopitem = new DesktopItem(buffer);

			desktopitemList.insert(desktopitemList.end(), desktopitem);

			Bang = "dummy ";
			Bang.append( xbuffer );
			Bang.append( " ");
			Bang.append( ybuffer );
			Bang.append( " \"" + (*it) + "\"");

			desktopitem->desktopload(hInstance, Bang.c_str());

			stringDesktopItemList.insert(stringDesktopItemList.end(), Bang);

			counter++;
			tempcounter++;
		}
	}
	//Update & Saves new Items, Only if:
	//	1. Some have changed
	else if (updateMode == -1)
	{
		StringListIterator sit;
		for ( sit = ActualDesktopItemPaths.begin(); sit != ActualDesktopItemPaths.end(); sit++)
		{
			bool found = false;

			for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
			{
				if (stricmp((*sit).c_str(), (*diit)->getIconCommand().c_str()) == NULL)
				{
					found = true;
					break;
				}
			}

			if (found)
			{
				int left = (*diit)->getX();
				int top = (*diit)->getY();

				if (left < 0)
					left = 0;
				if (top < 0)
					top = 0;
							
				string position = ( IntToString(left) + " " + IntToString(top) );

				Bang = ( "dummy " + position + " \"" + (*diit)->getIconCommand() + "\"");

				stringDesktopItemList.insert(stringDesktopItemList.end(), Bang);

				(*diit)->hide();
				desktopitemList.remove((*diit));
				delete *diit;
			}
			else
			{
				Bang = ( "dummy 100 100 \"" + (*sit) + "\"" );
			
				stringDesktopItemList.insert(stringDesktopItemList.end(), Bang);
			}
		}
		for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
		{
			(*diit)->hide();
			desktopitemList.remove((*diit));
			delete *diit;
		}

		desktopitemList.clear();
		
		int counter = 1;
		for( sit = stringDesktopItemList.begin(); sit != stringDesktopItemList.end(); sit++)
		{
			char buffer[32];
			strcpy(buffer, "xldi");
			char counterbuffer[16];
			itoa( counter, counterbuffer, 10 );
			if (counter < 10)
				strcat(buffer, "0");
			strcat(buffer, counterbuffer);

			char name[64], data[MAX_LINE_LENGTH];
			char *linebuffers[] = {name};

			LCTokenize((*sit).c_str(), linebuffers, 1, data);

			DesktopItem *desktopitem = new DesktopItem(name);

			desktopitemList.insert(desktopitemList.end(), desktopitem);

			desktopitem->desktopload(hInstance, (*sit));

			if (!useautodesktop)
			{
				Bang = ("@" + desktopSavePath + "@ @*xdesktopitem " + buffer + " " + data + "@");
				TextAppend(NULL, Bang.c_str());
			}
			counter++;
		}
	}
	//Update & Saves single changed Items, Only if:
	//	1. they have changed
	//	1. AutoCreateDesktop is ON
	else if (updateMode >= 20 && useautodesktop)
	{
		//Add Files
		if (updateMode == 21)
		{
			Bang = ( "dummy 100 100 \"" + updatedFile + "\"" );
		
			stringDesktopItemList.insert(stringDesktopItemList.end(), Bang);

			DesktopItem *desktopitem = new DesktopItem("dummy");

			desktopitemList.insert(desktopitemList.end(), desktopitem);

			desktopitem->desktopload(hInstance, Bang.c_str());
		}
		//Remove Files
		else if (updateMode == 22)
		{
			bool found = false;
			for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
			{
				if (stricmp((*diit)->getIconCommand().c_str(), updatedFile.c_str()) == NULL)
				{
					found = true;
					break;
				}
			}
			
			if (found)
			{
				char xname[32], xx[16], xy[16], xpath[MAX_PATH];
				char *buffers[] = {xname, xx, xy, xpath};

				for(StringListIterator sit = stringDesktopItemList.begin(); sit != stringDesktopItemList.end(); sit++)
				{
					LCTokenize((*sit).c_str(), buffers, 4, NULL);

					if (stricmp((*diit)->getIconCommand().c_str(), xpath) == NULL)
					{
						stringDesktopItemList.remove( *sit );
						break;
					}
				}

				(*diit)->hide();
				desktopitemList.remove((*diit));
				delete *diit;
			}
		}
		//Rename Files
		else if (updateMode == 23)
		{
			HANDLE hFindHandle;
			WIN32_FIND_DATA FindData;

			for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
			{
				if (strstr((*diit)->getIconCommand().c_str(), "::") == NULL)
				{
					hFindHandle = FindFirstFile((*diit)->getIconCommand().c_str(), &FindData);

					if(INVALID_HANDLE_VALUE == hFindHandle)
					{
						char xname[32], xx[16], xy[16], xpath[MAX_PATH];
						char *buffers[] = {xname, xx, xy, xpath};

						for(StringListIterator sit = stringDesktopItemList.begin(); sit != stringDesktopItemList.end(); sit++)
						{
							LCTokenize((*sit).c_str(), buffers, 4, NULL);

							if (stricmp((*diit)->getIconCommand().c_str(), xpath) == NULL)
							{
								stringDesktopItemList.remove( *sit );
								break;
							}
						}

						string positions = ( IntToString((*diit)->getX()) + " " + IntToString((*diit)->getY()) );

						(*diit)->hide();
						delete *diit;
						desktopitemList.remove( *diit );

						Bang = ( "dummy " + positions + " \"" + updatedFile + "\"" );

						//New Named DesktopItem
						stringDesktopItemList.insert(stringDesktopItemList.end(), Bang);

						DesktopItem *desktopitem = new DesktopItem("dummy");

						desktopitemList.insert(desktopitemList.end(), desktopitem);

						desktopitem->desktopload(hInstance, Bang.c_str());

						break;
					}
				}
			}

			FindClose(hFindHandle);
		}
	}
}

void DesktopSave(HWND hwndCaller, LPCSTR pszArgs)
{
	string internalcaller = pszArgs;

	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
	{
		if (stricmp((*diit)->getIconCommand().c_str(), internalcaller.c_str()) == NULL)
		{
			int left = (*diit)->getX();
			int top = (*diit)->getY();

			if (left < 0)
				left = 0;
			if (top < 0)
				top = 0;
						
			string position = ( IntToString(left) + " " + IntToString(top) );

			//Only Track change and do not save, if AutoCreate Desktop
			if (useautodesktop)
			{
				char xname[32], xx[16], xy[16], xpath[MAX_PATH];
				char *buffers[] = {xname, xx, xy, xpath};

				for(StringListIterator sit = stringDesktopItemList.begin(); sit != stringDesktopItemList.end(); sit++)
				{
					LCTokenize((*sit).c_str(), buffers, 4, NULL);

					if (stricmp((*diit)->getIconCommand().c_str(), xpath) == NULL)
					{
						Bang = ( (*diit)->getName() + " " + position + " \"" + (*diit)->getIconCommand() + "\"" );
							
						*sit = Bang;
							
						break;
					}
				}
			}
			//Save immediately, if no AutoCreate Desktop
			else
			{
				Bang = ("@" + desktopSavePath + "@ @[		]*(\\*[xX][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]).*(" + (*diit)->getName() + ")@ @\\1 \\2 " + position + " \"" + (*diit)->getIconCommand() + "\"@");
				if (!desktopSavePath.empty())
					TextReplace(NULL, Bang.c_str());
			}
		}
	}
}

// !xDesktopCreate <append>|<replace>|<update...> <path to lnkfolder>|<updatefile>
void BangCreate(HWND hwndCaller, LPCSTR pszArgs)
{
	StringList DesktopItemPaths;

	char action[32], path[MAX_LINE_LENGTH];
	char *buffers[] = {action, path};

	LCTokenize(pszArgs, buffers, 2, NULL);
	
	string searchPath = path;

	//Quit here, if no path is specified
	if (searchPath.empty())
		return;

	//Initialize
	bool defaultDesktop = false;
	
	if ( stricmp(searchPath.c_str(), ".desktop") == 0 || stricmp(searchPath.c_str(), "deskto") == 0 || stricmp(searchPath.c_str(), "eskto") == 0 )
		defaultDesktop = true;
	
	//Start Configuration and Writing to File
	if (!desktopSavePath.empty())
	{
		int updateMode;
		if ( stricmp(action, "update") == NULL )
			updateMode = -1;
		else if ( stricmp(action, "updateadd") == NULL )
			updateMode = 21;
		else if ( stricmp(action, "updatedel") == NULL )
			updateMode = 22;
		else if ( stricmp(action, "updateren") == NULL )
			updateMode = 23;
		else if (stricmp(action, "append") == NULL)
			updateMode = 1;
		else 
			updateMode = 0;

		if (updateMode == -1 || updateMode == 0 || updateMode == 1)
		{
			//Get All Items actual items
			DesktopItemPaths = getActualDesktopItemPaths(defaultDesktop, searchPath);

			//Clear all *xdesktopitem lines
			Bang = ("@" + desktopSavePath + "@ @\\*[xX][dD][eE][sS][kK][tT][oO][pP][iI][tT][eE][mM]@");
			TextDelete(NULL, Bang.c_str());

			if (updateMode == -1 || updateMode == 0)
			{
				stringDesktopItemList.clear();

				if (updateMode == 0)
				{
					Bang = "!xdesktopdestroy";
					LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
				}
			}
		}

		//0 = Create New (Replace/Reposition), 1 = Append (New will be Positioned), 2+ = Update
		DesktopUpdate(updateMode, searchPath, DesktopItemPaths);
	}
	else
	{
		Bang = "!alert \"You haven't defined the setting xDesktopConfigFile, which points to the file containing your *xDesktopItem lines!\"";
		LSExecute(NULL, Bang.c_str(), SW_SHOWNORMAL);
	}
}

// !xDesktopDestroy
void BangDestroy(HWND hwndCaller, LPCSTR pszArgs)
{
	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
		delete *diit;

	desktopitemList.clear();
}

// !xDesktopHide
void BangHide(HWND hwndCaller, LPCSTR pszArgs)
{
	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
		(*diit)->hide();
}

// !xDesktopShow
void BangShow(HWND hwndCaller, LPCSTR pszArgs)
{
	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
		(*diit)->show();
}

// !xDesktopToggle
void BangToggle(HWND hwndCaller, LPCSTR pszArgs)
{
	for(DesktopItemListIterator diit = desktopitemList.begin(); diit != desktopitemList.end(); diit++)
	{
		if((*diit)->isVisible())
			(*diit)->hide();
		else
			(*diit)->show();
	}
}

void getToken(char line[4096], char temp[4096], bool escape_codes)
{
    int offset = 0, i;
	bool endquote = false;
    char ENCAP = '@';

	while (line[offset] == ' ')
	{
		offset++;
	}

	if (line[offset] == ENCAP)
	{
		offset++;
		endquote = true;
	}

	for (i = 0; i+offset < int(strlen(line)); ++i)
	{
		// Copy else if block to add more escape codes.
		if (escape_codes && (line[i+offset] == '\\' && int(strlen(line)) != i+offset+1))
		{
		    if (line[i+offset+1] == ENCAP || line[i+offset+1] == '\\')
			{
				offset++;
				temp[i] = line[i+offset];
				continue;
			}
			// Bang escape code below. Present in 2.0 from Tony Chang
			else if (line[i+offset+1] == '^')
			{
				offset++;
				temp[i] = '!';
				continue;
			}
			// EVar escape code below. Present in 2.0 from Tony Chang
			else if (line[i+offset+1] == '#')
			{
				offset++;
				temp[i] = '$';
				continue;
			}
			// Comment escape code below. Added in 2.1l by leaf r&d
			else if (line[i+offset+1] == '~')
			{
				offset++;
				temp[i] = ';';
				continue;
			}
			// Quotation mark escape code below. Added in 2.2l by leaf r&d
			else if (line[i+offset+1] == '=')
			{
				offset++;
                temp[i] = '"';
                continue;
            }
		}

 	    if (line[i+offset] == ENCAP && endquote)
	  	    break;
		else if (line[i+offset] == ' ' && !endquote)
			break;
		else
			temp[i] = line[i+offset];
	}
	temp[i] = '\0';

	if (i+offset+1 > int(strlen(line)))
	{
		line[0] = '\0';
	}
	else
	{
		strcpy(line, line+i+offset+1);
	}
}

void TextAppend(HWND hwndCaller, LPCSTR pszArgs)
{
	char filename[4096], str[4096];
	char rest[4096];
	strcpy(rest, pszArgs);

	getToken(rest, filename, false);
	getToken(rest, str, false);

	ofstream file;
	file.open(filename, ios::app);

	if (!file.fail())
	{
		file << str << endl;
	}

	file.close();
}

void TextDelete(HWND hwndCaller, LPCSTR pszArgs)
{
	char filename[4096], regex[4096], input[4096];
	char rest[4096];
	regexp *r;
	strcpy(rest, pszArgs);

	getToken(rest, filename, false);
	getToken(rest, regex, true);

	r = regcomp(regex);

	ifstream file;
	file.open(filename);

	if (!file.fail())
	{
		list<string> l;
		string s;
		while (file.getline(input, 4096))
		{
			if (!regexec(r, input))
			{
				s = input;
				l.push_back(s);
			}
		}
		file.close();
		ofstream outfile;
		outfile.open(filename);
		for (list<string>::const_iterator lci = l.begin(); lci!=l.end(); lci++)
		{
			outfile << lci->data() << endl;
		}
		l.clear();
	}
}

void TextReplace(HWND hwndCaller, LPCSTR pszArgs)
{
	char filename[4096], regex[4096], input[4096], replace[4096];
	char rest[4096], buf[4096];
	regexp *r;
	strcpy(rest, pszArgs);

	getToken(rest, filename, false);
	getToken(rest, regex, true);
	getToken(rest, replace, true);

	r = regcomp(regex);

	ifstream file;
	file.open(filename);

	if (!file.fail())
	{
		list<string> l;
		string s;
		while (file.getline(input, 4096))
		{
			if (regexec(r, input))
			{
				regsub(r, replace, buf);
				s = buf;
				l.push_back(buf);
			}
			else
			{
				s = input;
				l.push_back(input);
			}
		}
		file.close();
		ofstream outfile;
		outfile.open(filename);
		for (list<string>::const_iterator lci = l.begin(); lci!=l.end(); lci++)
		{
			outfile << lci->data() << endl;
		}
		l.clear();
	}
}
